#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
# Univention Management Console
#  module: Open Xchange License Management
#
# Copyright (C) 2010-2014 Univention GmbH <http://www.univention.de/>
# by order of
#     Open-Xchange (http://www.open-xchange.com)
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

from univention.lib import Translation
from univention.management.console.modules import Base
from univention.management.console.log import MODULE
from univention.management.console.config import ucr
import univention.admin.uldap as udm_uldap
import univention.admin.uexceptions as udm_errors

import notifier.popen
import notifier.threads

import tempfile
import os, stat, traceback, types

import subprocess

import oxldbcore
from ldap import LDAPError

_ = Translation('univention-management-console-module-oxldb').translate

FN_LICENSE_BLACKLIST = '/usr/share/univention-management-console-module-oxldb/blacklisted_signatures'


class Instance( Base ):
	OPTION_MAPPING = ( ( 'username', 'ox/ldbaccount/username', '' ),
					   ( 'currentLicensekey', 'ox/licensekey', '' ),
					   ( 'ldapBase', 'ldap/base', ''),
					   ( 'serverRole', 'server/role', 'domaincontroller_master'),
					   )

	def __init__( self ):
		Base.__init__( self )

	def load( self, request ):
		"""Retrieve current configuration status of the Open Xchange system

		options: {}

		return: { <all OX related UCR variables> }
		"""
		result = {}
		ucr.load()
		for option, var, default in Instance.OPTION_MAPPING:
			result[ option ] = ucr.get( var, default )

		username = ucr.get('ox/ldbaccount/username')
		password = ''
		pwd_file = ucr.get('ox/ldbaccount/passwordfile')
		result['passwordExistsInBackend'] = bool(pwd_file and os.path.exists(pwd_file))
		if result['passwordExistsInBackend']:
			try:
				password = open(pwd_file,'r').read().strip('\n\r')
			except IOError:
				password = ''
		result['credentialsValid'] = False

		def _thread( request ):
			MODULE.info( 'trying to read current UCS license type' )
			license_types = self._get_ucs_license_types()
			MODULE.info( 'license_types = %r' % (license_types,))
			MODULE.info( 'trying to testconnect to LDB with username=%s' % username )
			oxldb = oxldbcore.OXLDB()
			connect_result = oxldb.ldb_connect(username=username, password=password)
			MODULE.info( 'connect_result %s' % (connect_result))
			result['credentialsValid'] = (connect_result == True)
			return result

		if username and password:
			thread = notifier.threads.Simple('licensekey_get', notifier.Callback(_thread, request), notifier.Callback(self._thread_finished, request))
			thread.run()
		else:
			self.finished( request.id, result )

	def _check_thread_error( self, thread, result, request ):
		"""Checks if the thread returned an exception. In that case in
		error response is send and the function returns True. Otherwise
		False is returned."""
		if not isinstance( result, BaseException ):
			return False

		msg = '%s\n%s: %s\n' % ( ''.join( traceback.format_tb( thread.exc_info[ 2 ] ) ), thread.exc_info[ 0 ].__name__, str( thread.exc_info[ 1 ] ) )
		MODULE.process( 'An internal error occurred: %s' % msg )
		self.finished( request.id, None, msg, False )
		return True

	def _thread_finished( self, thread, result, request ):
		if self._check_thread_error( thread, result, request ):
			return
		self.finished( request.id, result )

	def _get_ucs_license_types(self):
		ldap_server = ucr.get('ldap/server/name')
		ldap_base = ucr.get( 'ldap/base' )
		try:
			lo = udm_uldap.access( host = ldap_server,
								   base = ldap_base,
								   binddn = self._user_dn,
								   bindpw = self._password,
								   start_tls = 2 )
		except udm_errors.noObject, e:
			MODULE.error('cannot establish LDAP connection to server %s' % (ldap_server,))
			return []
		except LDAPError, e:
			MODULE.error('cannot establish LDAP connection to server %s' % (ldap_server,))
			return []

		try:
			result = lo.search(filter='(&(objectClass=univentionLicense)(cn=admin))',
								base=ldap_base, scope='sub',
								attr=['univentionLicenseType', 'univentionLicenseSignature'])
		except LDAPError, e:
			MODULE.error('looking for UCS license object failed: %s' % (ldap_server,))
			return []

		# load blacklisted license signatures
		try:
			blacklisted_signatures = open(FN_LICENSE_BLACKLIST, 'r').read().splitlines()
		except (IOError, OSError), e:
			MODULE.error('cannot read list of blacklisted hashes: %s' % (e,))
			return []

		if result:
			# at least one license object has been found
			for dn, attrs in result:
				license_signature = attrs.get('univentionLicenseSignature',[None])[0]
				if license_signature and not license_signature in blacklisted_signatures:
					# signature is valid
					return attrs.get('univentionLicenseType',[])
		return []

	def licensekeys_get( self, request ):
		"""Check given LDB credentials and return a list of tuples containing OX licensekey and corresponding ldap base.

		options:
			username: LDB username
			password: LDB password
			passwordExistsInBackend: this boolean flag indicates if the original password has been unchanged and is known in backend

		return { 'success' : (True|False), licensekeys: [ { 'id': <LICENSEKEY>, 'label': <LICENSEKEY>, ldapBase: <LDAPBASE>, eulaURL: <URL2EULA> }, ... ] }
		"""
		def _thread( request ):
			result = { 'success': False,
					   'licensekeys': [],
					   'message': _('Connection to OX LDB failed!'),
					   }

			username = request.options.get( 'username' )
			password = request.options.get( 'password' )
			if request.options.get( 'passwordExistsInBackend' ):
				pwd_file = ucr.get( 'ox/ldbaccount/passwordfile' )
				if pwd_file and os.path.exists(pwd_file):
					try:
						password = open(pwd_file,'r').read().strip('\n\r')
					except IOError:
						password = ''

			MODULE.info('trying to connect to LDB with username=%s' % username)
			oxldb = oxldbcore.OXLDB()
			connect_result = oxldb.ldb_connect(username=username, password=password)
			MODULE.info('connect_result %s' % (connect_result))
			if connect_result == True:
				oxldb.username = username
				oxldb.password = password
				result['success'] = True
				result['message'] = _('Credentials have been tested successfully.')

				# get all license keys
				licensekeys = oxldb.get_keys(products=oxldb.detected_products)
				MODULE.info( 'licensekeys %s' % ( licensekeys ) )

				# request additional data for each license key
				for licensekey in licensekeys:
					MODULE.info( 'licensekey %s' % ( licensekey ) )

					# check if EULA has been accepted and if EULA URL is available
					eula_url = oxldb.ldb_eula_get_url(licensekey=licensekey)
					MODULE.info( 'eula_url %s' % ( eula_url ) )
					if eula_url and eula_url.lower() != "none":
						eula_accepted = oxldb.ldb_eula_has_been_accepted(licensekey=licensekey)
						MODULE.info( 'eula_accepted %s' % ( eula_accepted ) )
						if eula_accepted:
							eula_url = ''
					else:
						eula_url = ''

					# get license and determine ldap base
					licensedata = oxldb.ldb_get_license_and_ldapbase(licensekey=licensekey)
					MODULE.info( 'licensedata %s' % ( licensedata ) )

					# if no license is available, then return empty values
					if not licensedata:
						licensedata = { 'license': None, 'ldapbase': None }

					# append new entry to result
					result['licensekeys'].append( { 'id': licensekey,
													'label': licensekey,
													'eulaURL': eula_url,
													'licenseAvailable': bool(licensedata.get('license','')),
													'ldapBase': licensedata.get('ldapbase') } )
			else:
				result['message'] = _('Connection to OX LDB failed: %s') % connect_result
			MODULE.info( 'result %s' % ( result ) )
			return result

		thread = notifier.threads.Simple( 'licensekey_get', notifier.Callback( _thread, request ),
										  notifier.Callback( self._thread_finished, request ) )
		thread.run()

	def save( self, request ):
		"""Save given LDB credentials, import UCS license and update repository entries

		options:
			username: LDB username
			password: LDB password
			passwordExistsInBackend: this boolean flag indicates if the original password has been unchanged and is known in backend
			licensekey: OX license key
			eulaAccepted: EULA has been accepted

		return { 'success' : (True|False), 'message': <string> }
		"""
		def _thread( request ):
			result = { 'success': True,
					   'message': '',
					   }

			oxldb = oxldbcore.OXLDB()

			username = request.options.get( 'username' )
			password = request.options.get( 'password' )
			if request.options.get( 'passwordExistsInBackend' ):
				pwd_file = ucr.get( oxldb.UCR_OX_LDBACCOUNT_PASSWDFN )
				if pwd_file and os.path.exists(pwd_file):
					try:
						password = open(pwd_file,'r').read().strip('\n\r')
					except IOError:
						password = ''
			licensekey = request.options.get( 'licensekey' )
			eulaAccepted = request.options.get( 'eulaAccepted' )

			MODULE.info('trying to connect to LDB with username=%s' % username)
			connect_result = oxldb.ldb_connect(username=username, password=password)
			MODULE.info('connect_result %s' % ( result ))
			if connect_result == True:
				oxldb.username = username
				oxldb.password = password

				# update credentials
				MODULE.info('setting username...')
				msg = oxldb.set_username(username)
				if not msg:
					MODULE.info('setting password...')
					msg = oxldb.set_password(password)
				MODULE.info('updating components...')
				oxldb.update_components()
				if msg:
					result['message'] += _('<li>Failed to set username or password: %s</li>') % str(msg)
					result['success'] = False
					MODULE.error('Failed to set username/password: %s' % str(msg))
				else:
					result['message'] += '<li>OX LDB account has been set successfully.</li>'

				# update license key
				MODULE.info('fetching licensekeys...')
				licensekeys = oxldb.get_keys(products=oxldb.detected_products)
				MODULE.info('licensekeys %s' % ( licensekeys ))
				if licensekey in licensekeys:
					# check if product type changed by comparing old and new licensekey
					producttype_old = oxldb.get_product_type(ucr.get(oxldb.UCR_OX_LICENSEKEY,''))
					producttype_new = oxldb.get_product_type(licensekey)
					MODULE.info('old product type = %s      new product type = %s' % (ucr.get(oxldb.UCR_OX_LICENSEKEY,''), licensekey))
					MODULE.info('old product type = %s      new product type = %s' % (producttype_old, producttype_new))

					if producttype_old and producttype_new and producttype_old != producttype_new:
						result['note'] = _('The product type has changed from %s to %s. It is advisable to clear the browser cache to prevent layout difficulties.') % (producttype_old.upper(), producttype_new.upper())

					# if licensekey
					msg = oxldb.ldb_set_license_key(licensekey)
					MODULE.info('set_license_key(%s) = %s' % (licensekey, msg))
					if msg is not True:
						result['message'] += _('<li>Failed to set license key: %s</li>') % str(msg)
						result['success'] = False
					else:
						result['message'] += _('<li>License key has been set successfully.</li>')
				else:
					MODULE.error('cannot find licensekey %s in LDB' % licensekey)
					result['message'] += _('<li>Failed to select license key: license key not found in LDB</li>')
					result['success'] = False

				# accept eula
				if eulaAccepted:
					MODULE.info('user accepted EULA')
					msg = oxldb.ldb_eula_accept(licensekey)
					MODULE.info('ldb_eula_accept(%s) = %s' % (licensekey, msg))
					if msg is not True:
						result['message'] += _('<li>Failed to accept EULA: %s</li>') % str(msg)
						result['success'] = False
					else:
						result['message'] += _('<li>EULA has been accepted successfully.</li>')

				# import license
				if oxldb.is_oxse():
					MODULE.info('OXSE system detected - license import skipped')
				else:
					licensedata = oxldb.ldb_get_license_and_ldapbase(licensekey=licensekey)
					if licensedata:
						licensestring = licensedata.get('license')
						ldapbase = licensedata.get('ldapbase','')
					else:
						licensestring = None
						ldapbase = None

					# import license only if ldap base matches
					if licensestring in (None, '', 'None'):
						MODULE.warn('skipping license import - license is empty')
						result['message'] += _('<li>The import of the UCS license has been skipped due to missing UCS license for this license key.</li>')
					elif ldapbase is None or ldapbase.strip() != ucr.get('ldap/base'):
						MODULE.error('skipping license import due to mismatching LDAP base')
						MODULE.error('ldap base of license: %s' % ldapbase)
						MODULE.error('ldap base of system: %s' % ucr.get('ldap/base'))
						result['message'] += _('<li>The import of the UCS license has been skipped due to mismatching LDAP base.</li>')
					else:

						MODULE.info('licensestring:\n%s' % repr(licensestring))
						if not licensestring in ('', 'None', None):
							# write license to tempfile
							try:
								tmpfn = tempfile.mkstemp( '.ldif', 'ucs-license.')[1]
								os.chmod(tmpfn, stat.S_IRUSR | stat.S_IWUSR)
								open(tmpfn, 'w').write(str(licensestring))
								os.chmod(tmpfn, stat.S_IRUSR)
							except (OSError, IOError), e:
								MODULE.error('cannot write license: %s' % str(e))
								result['message'] += _('<li>Failed to write UCS license to disk: %s</li>') % str(e)
								result['success'] = False

							# call univention-license-import
							try:
								logfn = tempfile.mkstemp( '.log', 'univention-license-import')[1]
								exitcode = subprocess.call( 'univention-license-import %s > %s 2> %s' % (tmpfn, logfn, logfn) , shell=True)
								MODULE.info('univention-license-import returned exitcode %s' % exitcode)

								errormsg = ''
								lines = open(logfn,'r').readlines()
								MODULE.info('Output of univention-license-import:\n%s' % lines)
								for line in lines:
									if line.startswith('E: '):
										errormsg += line[3:] + '<br />'
								if errormsg:
									MODULE.error('failed to import license:')
									MODULE.error('\n'.join(lines))
									result['message'] += _('<li>Failed to import UCS license:<br>%s</li>') % errormsg
									result['success'] = False

							except Exception, e:
								MODULE.error('cannot import license: %s' % str(e))
								result['message'] += _('<li>UCS license update has failed - please try again (later) or contact Open-Xchange</li>')
								result['success'] = False

							# cleanup temporary files
							for fn in (tmpfn, logfn):
								if os.path.exists(fn):
									try:
										os.remove(fn)
									except Exception, e:
										pass
							oxldb.update_components()

				if result['success']:
					result['message'] = _('<p>The following steps have been completed:<ul>%s</ul></p>') % result['message']
				else:
					result['message'] = _('<p>While updating settings some steps failed:<ul>%s</ul></p>') % result['message']

			else:
				result['message'] += _('<li>Connection to OX LDB failed - please try again (later)</li>')
				result['success'] = False
			MODULE.info( 'result %s' % ( result ) )
			return result

		thread = notifier.threads.Simple( 'save', notifier.Callback( _thread, request ),
										  notifier.Callback( self._thread_finished, request ) )
		thread.run()

	def reset_password( self, request ):
		"""Reset given LDB credentials

		options:
			username: LDB username
			password: LDB password

		return { 'success' : (True|False), 'message': <string> }
		"""

		def _thread( request ):
			result = { 'success': False,
					   'message': '',
					   }

			username = request.options.get( 'username' )
			password = request.options.get( 'password' )

			if username and password:
				# username and password are "valid" so start password reset
				oxldb = oxldbcore.OXLDB()
				value = oxldb.ldb_reset_password(username, password)
				if type(value) is types.BooleanType:
					if value:
						MODULE.info('password reset for username "%s" has been successful' % username)
						result['message'] = _('Your new password has been accepted. In order to complete the password reset request, an email is sent to you. The mail contains a link that you have to open in your browser to confirm the password reset.')
						result['success'] = True
					else:
						MODULE.error('password reset returned False')
						result['message'] = _('An error occured. Please try again later.')
				else:
					MODULE.error('password reset failed: %s' % str(value))
					result['message'] = _('An error occured during password reset:<br />') + str(value)
			else:
				MODULE.error('No valid username and/or password given!')
				result['message'] = _('No valid username and/or password given!')

			MODULE.info( 'result %s' % ( result ) )
			return result

		thread = notifier.threads.Simple( 'passwordreset', notifier.Callback( _thread, request ),
										  notifier.Callback( self._thread_finished, request ) )
		thread.run()
