#!/usr/bin/python2.7
#
# Copyright (C) 2010-2014 Univention GmbH <http://www.univention.de/>
# by order of
#     Open-Xchange (http://www.open-xchange.com)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import univention.management.console as umc

import suds, types, re, hashlib, urllib2, traceback, os, stat, subprocess, copy, time
from univention.management.console.log import MODULE
import univention.config_registry

_ = umc.Translation( 'univention-management-console-module-oxldb' ).translate


class OXLDB(object):
	UCR_ALLOWED_CHARACTERS = '^[^#:@]+$'
	OX_ALLOWED_CHARACTERS_USERNAME = '^.+$'  # allow all characters but at least one is required
	RE_ALLOWED_DEBIAN_PKGNAMES = re.compile('^[a-z0-9][a-z0-9.+-]+$')
	FN_UPDATER_APTSOURCES_COMPONENT = '/etc/apt/sources.list.d/20_ucs-online-component.list'
	FN_OX_LDB_PASSWORD = '/etc/ox-secrets/ldb-account.secret'
	URL_OX_WSDL_HTTP  = 'http://ldb.open-xchange.com/licenseDBServer/services/LdbService?wsdl'
	URL_OX_WSDL_HTTPS = 'https://ldb.open-xchange.com/licenseDBServer/services/LdbService?wsdl'
	UCR_OX_LDB_FORCE_PROTOCOL = 'ox/ldb/force/protocol'
	UCR_OX_LDBACCOUNT_USER = 'ox/ldbaccount/username'
	UCR_OX_LDBACCOUNT_PASSWDFN = 'ox/ldbaccount/passwordfile'
	UCR_OX_COMPONENTS_AVAILABLE = 'ox/components/available'
	UCR_OX_LICENSEKEY = 'ox/licensekey'
	COMPONENT_DISABLED = 'disabled'
	COMPONENT_AVAILABLE = 'available'
	COMPONENT_PERMISSION_DENIED = 'permission denied'
	COMPONENT_UNKNOWN = 'unknown'
	# OX Keys:
	# OX6-SE4UCS-MK-
	# OX6-SE4UCS-SUB-
	# OX6-SE-MK-
	# OX6-SE-SUB-
	# OX-APPS-SUB-
	# OX-DOC-SUB-
	OX_KEY_REGEX = { 'oxae' : re.compile('^OX6-DEPRECATED'),
					 'oxase' : re.compile('(^OX6-AS?E(-.*?)?-(MK|SUB)-|^[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}$)'),
					 'oxse' : re.compile('(^OX6?-(SE(4UCS)?|APPS|DOC)(-.*?)?-(MK|SUB)-|^[0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12}$)'),
					 }

	def __init__(self, connect=False):
		self.configRegistry = univention.config_registry.ConfigRegistry()
		self.configRegistry.load()
		self.suds = None
		self.suds_sessionId = None
		self.keys_all = []
		self.detected_products = ['oxse']

		self.licensekey = self.configRegistry.get(self.UCR_OX_LICENSEKEY, '').strip()
		self.username = self.configRegistry.get(self.UCR_OX_LDBACCOUNT_USER,'').strip()
		self.password_fn = self.configRegistry.get(self.UCR_OX_LDBACCOUNT_PASSWDFN, self.FN_OX_LDB_PASSWORD).strip()
		self.credentials_valid = False
		if not self.configRegistry.get(self.UCR_OX_LDBACCOUNT_PASSWDFN, '').strip():
			try:
				univention.config_registry.handler_set( [ u'%s=%s' % (self.UCR_OX_LDBACCOUNT_PASSWDFN, self.password_fn ) ] )
			except:
				pass

		try:
			self.password = open(self.password_fn,'r').read().strip('\n\r')
		except:
			self.password = self.guess_password()

		if connect:
			result = self.ldb_connect()
			if not result == True:
				MODULE.error('OXLDB: __init__: connect: %s' % str(result))

	def guess_password(self):
		if not self.username:
			return ''

		components = self.configRegistry.get(self.UCR_OX_COMPONENTS_AVAILABLE,'').split(',')
		for component in components:
			if component:
				if self.configRegistry.get('repository/online/component/%s' % component,'').lower() in ('yes','true','enable','enabled'):
					if self.configRegistry.get('repository/online/component/%s/username' % component) == self.username:
						passwd = self.configRegistry.get('repository/online/component/%s/password' % component)
						if passwd:
							return passwd
		return ''

	def mapType(self, result):
		if type(result) is types.BooleanType:
			return result
		elif type(result) is suds.sax.text.Text:
			return str(result.encode('utf-8'))
		elif type(result) is types.NoneType:
			return result
		return result

	def get_keys(self, products=None, getall=False):
		"""
		Returns list of ox maintenance keys. If all==True, then all available keys are returned.
		If 'products' is empty, then all keys for detected ox 'products' are returned.
		If 'products' contains list of ox products, only those keys are returned.
		"""
		if getall:
			return self.keys_all

		if products:
			prodlist = copy.deepcopy(products)
		else:
			prodlist = [ 'oxse', 'oxase', 'oxae' ]

		result = []
		for ver in prodlist:
			result.extend( [ x for x in self.keys_all if self.OX_KEY_REGEX[ver].match(x) ] )

		return list(set(result))

	def get_product_type(self, licensekey):
		for key, regex in self.OX_KEY_REGEX.items():
			if regex.match(licensekey):
				return key
		return None

	def is_oxae(self):
		return 'oxae' in self.detected_products

	def is_oxse(self):
		return 'oxse' in self.detected_products

	def is_oxase(self):
		return 'oxase' in self.detected_products

	def is_ox(self):
		return bool(self.detected_products)

	def get_productname(self):
		if 'oxse' in self.detected_products:
			return _('Open-Xchange Server Edition for UCS')
		elif 'oxase' in self.detected_products:
			return _('Open-Xchange Advanced Server Edition')
		elif 'oxae' in self.detected_products:
			return _('Open-Xchange Appliance Edition')
		else:
			return _('OX')

	def get_productname_short(self):
		if 'oxse' in self.detected_products:
			return _('OX SE4UCS')
		elif 'oxase' in self.detected_products:
			return _('OX ASE')
		elif 'oxae' in self.detected_products:
			return _('OX AE')
		else:
			return _('OX')

	def is_account_complete(self):
		return self.username and self.password

	def is_account_valid(self):
		return self.credentials_valid

	def ldb_connect(self, username=None, password=None):
		"""
		Connect to LDB via SOAP
		return value:
			True     ==> successful connected
			<string> ==> error message
		"""
		# use given credentials or as fallback the known ones
		user = self.username
		if username:
			user = username
		passwd = self.password
		if password:
			passwd = password

		# if still connected then disconnect first
		if self.suds:
			self.ldb_disconnect()

		# stop here if credentials are missing
		if not user or not passwd:
			MODULE.error('OXLDB: ldb_connect: credentials missing')
			return _('missing credentials')

		proto_list = { 'http': self.URL_OX_WSDL_HTTP,
					   'https': self.URL_OX_WSDL_HTTPS,
					   }

		# setup connection
		try:
			proxy = self.configRegistry.get('proxy/http','')
			if proxy:
				MODULE.info('OXLDB: ldb_connect: using proxy %s' % proxy)
				proxysettings = dict( http=proxy, https=proxy )
				protocol = self.configRegistry.get(self.UCR_OX_LDB_FORCE_PROTOCOL,'http')
				MODULE.info('OXLDB: ldb_connect: using protocol %s' % protocol)
				wsdl = proto_list[protocol]
				MODULE.info('OXLDB: ldb_connect: using wsdl %s' % wsdl)
				self.suds = suds.client.Client(wsdl, proxy=proxysettings )
			else:
				MODULE.info('OXLDB: ldb_connect: no proxy')
				protocol = self.configRegistry.get(self.UCR_OX_LDB_FORCE_PROTOCOL,'https')
				MODULE.info('OXLDB: ldb_connect: using protocol %s' % protocol)
				wsdl = proto_list[protocol]
				MODULE.info('OXLDB: ldb_connect: using wsdl %s' % wsdl)
				self.suds = suds.client.Client(wsdl)
			MODULE.info('OXLDB: ldb_connect: got suds object')
			cache = self.suds.options.cache
			cache.setduration(minutes=30)
			MODULE.info('OXLDB: ldb_connect: cache timeout set')
			# test if service is available
			available = self.mapType(self.suds.service.checkAvailable())
		except urllib2.URLError, e:
			MODULE.error('OXLDB: ldb_connect: setup: %s' % str(e))
			MODULE.info('OXLDB: ldb_connect: setup: %s' % traceback.format_exc())
			return _('connection error: %r') % (e.args,)
		except Exception, e:
			self.suds = None
			MODULE.error('OXLDB: ldb_connect: setup: %s' % str(e))
			MODULE.info('OXLDB: ldb_connect: setup: %s' % traceback.format_exc())
			if type(e.args) == type(()) and type(e.args[0]) == type(()) and len(e.args[0]) > 1:
				return _('error %s: %s') % (e.args[0][0], e.args[0][1])
			return _('service failure - please try again (later)')

		if not available:
			self.suds = None
			MODULE.error('OXLDB: ldb_connect: service unavailable')
			return _('service unavailable - check connection')

		# get sessionid
		try:
			self.suds_sessionId = self.mapType(self.suds.service.login(username=user, passwordMd5=hashlib.md5(passwd).hexdigest()))
		except Exception, e:
			if 'Invalid credentials' in str(e):
				self.suds = None
				self.credentials_valid = False
				MODULE.error('OXLDB: ldb_connect: login: %s' % str(e))
				return _('invalid credentials')
			else:
				self.suds = None
				MODULE.error('OXLDB: ldb_connect: login: %s' % str(e))
				return _('service failure - please try again (later)')
		self.credentials_valid = True

		# get all license keys
		try:
			self.keys_all = self.suds.service.getAllKeysForThisAccount(sessionId=self.suds_sessionId)
		except Exception, e:
			self.suds = None
			self.suds_sessionId = None
			MODULE.error('OXLDB: ldb_connect: keys: %s' % str(e))
			return _('service failure - please try again (later)')

		return True

	def ldb_disconnect(self):
		"""
		disconnect from LDB service
		"""
		if self.suds:
			try:
				self.suds.service.logout(sessionId=self.suds_sessionId)
			except Exception, e:
				MODULE.error('OXLDB: ldb_disconnect: logout failed: %s' % str(e))
			self.suds = None
			self.suds_sessionId = None
		return True

	def ldb_reconnect(self):
		"""
		reconnect to LDB service
		return value:
			see ldb_connect()
		"""
		self.ldb_disconnect()
		return self.ldb_connect()

	def ldb_is_connected(self):
		"""
		check if connection to LDB service is available
		return value:
			<bool>	==> status 'ldb connection exists'
		"""
		return bool(self.suds)

	def ldb_eula_has_been_accepted(self, licensekey=None):
		"""
		fetch acceptance state of EULA for specified license key
		return value:
			None     ==> ERROR
			True     ==> EULA has been accepted
			False    ==> EULA has not been accepted
		"""
		lkey = licensekey
		if not lkey:
			lkey = self.licensekey

		# doing a reconnect is a workaround for broken LDB; theoretically not required
		#self.ldb_reconnect()
		if not self.ldb_is_connected():
			MODULE.error('OXLDB: ldb_eula_has_been_accepted: not connected')
			return None
		try:
			state = self.mapType(self.suds.service.isEulaAccepted(sessionId=self.suds_sessionId, licenseKey=lkey))
			MODULE.info('OXLDB: ldb_eula_has_been_accepted: state=%s' % repr(state))
		except Exception, e:
			MODULE.error('OXLDB: ldb_eula_has_been_accepted: %s' % str(e))
			return None

		return state

	def ldb_eula_get_url(self, licensekey=None):
		"""
		fetch URL of EULA for specified license key
		return value:
			None     ==> Error
			<string> ==> URL
		"""
		lkey = licensekey
		if not lkey:
			lkey = self.licensekey

		# doing a reconnect is a workaround for broken LDB; theoretically not required
		#self.ldb_reconnect()
		if not self.ldb_is_connected():
			MODULE.error('OXLDB: ldb_eula_get_url: not connected')
			return None
		try:
			url = self.mapType(self.suds.service.getEulaUrl(sessionId=self.suds_sessionId, licenseKey=lkey))
			MODULE.info('OXLDB: ldb_eula_get_url: url=%s' % repr(url))
		except Exception, e:
			MODULE.error('OXLDB: ldb_eula_get_url: %s' % str(e))
			return None

		if url == None:
			url = ''
		return str(url)

	def ldb_eula_accept(self, licensekey=None):
		"""
		accept EULA for specified license key
		return value:
			True     ==> success
			<string> ==> error message
		"""
		lkey = licensekey
		if not lkey:
			lkey = self.licensekey

		# doing a reconnect is a workaround for broken LDB; theoretically not required
		#self.ldb_reconnect()
		if not self.ldb_is_connected():
			return None
		try:
			state = self.mapType(self.suds.service.acceptEula(sessionId=self.suds_sessionId, licenseKey=lkey))
			MODULE.info('OXLDB: ldb_accept_eula: state=%s' % repr(state))
		except Exception, e:
			MODULE.error('OXLDB: ldb_accept_eula: %s' % str(e))
			return str(e)

		return True

	def ldb_get_license(self, licensekey=None):
		"""
		fetch UCS license file for specified license key
		return value:
			<string> ==> LDIF
			None     ==> error
		"""
		lkey = licensekey
		if not lkey:
			lkey = self.licensekey

		# doing a reconnect is a workaround for broken LDB; theoretically not required
		#self.ldb_reconnect()
		if not self.ldb_is_connected():
			return None
		try:
			data = self.mapType(self.suds.service.getProductLicense(sessionId=self.suds_sessionId, licenseKey=lkey))
		except Exception, e:
			MODULE.error('OXLDB: ldb_get_license: %s' % str(e))
			return None

		return str(data)

	def ldb_get_license_and_ldapbase(self, licensekey=None):
		"""
		fetch UCS license file for specified license key and guess ldap base
		return value:
		    { 'license': <string>, 'ldapbase': <string> } ==> OK
            None ==> error
		"""
		licensestring = self.ldb_get_license(licensekey)
		if not licensestring:
			return None
		ldapbase = None
		for line in licensestring.splitlines():
			if line.startswith('univentionLicenseBaseDN: '):
				ldapbase = line.split(': ',1)[1]
		return { 'license': licensestring, 'ldapbase': ldapbase }

	def ldb_set_license_key(self, licensekey):
		"""
		Write specified licensekey to UCR variable
		return value:
			True     ==> success
			<string> ==> error message
		"""
		try:
			univention.config_registry.handler_set( [ u'%s=%s' % (self.UCR_OX_LICENSEKEY, licensekey ) ] )
			self.licensekey = licensekey
		except Exception, e:
			MODULE.error('OXLDB: ldb_set_license_key: %s' % str(e))
			return _('setting license key failed - please try again (later)')

		return True

	def ldb_reset_password(self, username, password):
		"""
		Send password reset request for specified username. The specified password has been given by user.
		return value:
			True     ==> success
			<string> ==> error message
		"""
		# setup connection and reset password
		try:
			proxy = self.configRegistry.get('proxy/http','')
			if proxy:
				MODULE.info('OXLDB: ldb_reset_password: using proxy %s' % proxy)
				proxysettings = dict( http=proxy, https=proxy )
				wsdl = self.URL_OX_WSDL_HTTP
				c = suds.client.Client(wsdl, proxy=proxysettings )
			else:
				MODULE.info('OXLDB: ldb_connect: no proxy')
				wsdl = self.URL_OX_WSDL_HTTPS
				c = suds.client.Client(wsdl)
			MODULE.info('OXLDB: ldb_connect: got suds object')
			result = self.mapType(c.service.requestPasswordChange(username=username, newPassMd5=hashlib.md5(password).hexdigest()))
		except urllib2.URLError, e:
			MODULE.error('OXLDB: ldb_connect: setup: %s' % str(e))
			MODULE.info('OXLDB: ldb_connect: setup: %s' % traceback.format_exc())
			return _('password reset failed - error %s: %s') % (e.args[0][0], e.args[0][1])
		except Exception, e:
			MODULE.error('OXLDB: ldb_reset_password: %s' % str(e))
			return _('password reset failed - please try again (later)')
		return result

	def clear_credentials(self):
		msg = None
		try:
			univention.config_registry.handler_unset( [ u'%s' % self.UCR_OX_LDBACCOUNT_USER, u'%s' % self.UCR_OX_LDBACCOUNT_PASSWDFN ] )
			self.username = None
			self.password = None
		except:
			msg = _('Cannot reset username and password.')
		return msg

	def set_username(self, username):
		msg = None
		try:
			univention.config_registry.handler_set( [ u'%s=%s' % (self.UCR_OX_LDBACCOUNT_USER, username ) ] )
			self.username = username
		except:
			msg = _('Cannot write username to UCR variable.')
		return msg

	def set_password(self, password):
		self.configRegistry.load()
		self.password_fn = self.configRegistry.get(self.UCR_OX_LDBACCOUNT_PASSWDFN, self.FN_OX_LDB_PASSWORD).strip()

		msg = None
		if os.path.exists(self.password_fn) and not os.path.isfile(self.password_fn):
			msg = _('Cannot write password into file %s') % self.password_fn
			MODULE.error('OXLDB: %s exists and is no file' % self.password_fn)
		else:
			try:
				open(self.password_fn,'w+').write( password )
				os.chmod(self.password_fn, stat.S_IRUSR)
				self.password = password
			except:
				MODULE.error('OXLDB: Cannot write password into file %s:\n%s' % (self.password_fn, traceback.format_exc().replace('%','#')) )
				msg = _('Cannot write password into file %s') % self.password_fn

		return msg

	def update_components(self):
		setlist = []
		components = self.configRegistry.get(self.UCR_OX_COMPONENTS_AVAILABLE,'').split(',')
		for component in components:
			if component:
				setlist.append( u'%s=%s' % ('repository/online/component/%s/username' % component, self.username) )
				setlist.append( u'%s=%s' % ('repository/online/component/%s/password' % component, self.password) )
		if setlist:
			try:
				univention.config_registry.handler_set( setlist )
			except:
				pass

	def get_component_list(self):
		"""
		get status of all known OX components
		return value:
			{ <componentname>: [ <description>, <COMPONENT_AVAILABLE/COMPONENT_PERMISSION_DENIED/COMPONENT_DISABLED/COMPONENT_UNKNOWN> ],
			  ...
			  }
		"""
		components = {}
		for component in self.configRegistry.get(self.UCR_OX_COMPONENTS_AVAILABLE,'').split(','):
			components[component] = { 'description': None,
									  'status': self.COMPONENT_UNKNOWN,
									  }
			components[component]['description'] = self.configRegistry.get('repository/online/component/%s/description' % component)
			if component and self.configRegistry.get('repository/online/component/%s' % component):
				if not self.configRegistry.get('repository/online/component/%s' % component,'disabled').lower() in ('enabled','true','yes','1'):
					components[component]['status'] = self.COMPONENT_DISABLED
				else:
					try:
						lines = open(self.FN_UPDATER_APTSOURCES_COMPONENT,'r').readlines()
					except:
						lines = []
					for line in lines:
						if not line.startswith('#') and ('/maintained/component/%s/' % component) in line:
							components[component]['status'] = self.COMPONENT_AVAILABLE
						elif line.startswith('# Authentication failure for') and ('/maintained/component/%s/' % component) in line:
							components[component]['status'] = self.COMPONENT_PERMISSION_DENIED
			else:
				components[component]['status'] = self.COMPONENT_DISABLED
			if not component in components:
				components[component]['status'] = self.COMPONENT_UNKNOWN
		return components

	def get_component_defaultpackage(self, component):
		"""
		returns a list of (meta) package names to be installed for this component
		return value:
			[ <string>, ... ]
		"""
		lst = []
		if component and self.configRegistry.get('repository/online/component/%s/defaultpackage' % component):
			val = self.configRegistry.get('repository/online/component/%s/defaultpackage' % component, '')
			lst = val.split(' ')
		return lst

	def is_component_defaultpackage_installed(self, component):
		"""
		returns installation status of component's default packages
		return value:
			True  ==> all default packages are installed
			False ==> at least one package is not installed
		"""
		pkglist = self.get_component_defaultpackage(component)
		result = True
		if pkglist:
			for pkg in pkglist:
				match = self.RE_ALLOWED_DEBIAN_PKGNAMES.search(pkg)
				if match:
					exitcode = subprocess.call( 'dpkg -s %s | grep "^Status: install ok installed"' % pkg, shell=True )
					if exitcode:
						result = False
				else:
					MODULE.error('OXLDB: is_component_defaultpackage_installed:  invalid package name (%s)' % (pkg))
		return result
